from functools import cache
from pathlib import Path
from typing import Iterable

import cv2
import numpy as np
from PIL import Image


@cache
def read_image_lab(path: Path) -> np.ndarray:
    """Read image and convert to LAB color space in CHW format with values 0-255"""
    img = cv2.imread(str(path))

    assert img is not None, f"Could not read image from {path}"

    # Convert BGR (OpenCV default) → LAB
    img = cv2.cvtColor(img, cv2.COLOR_BGR2LAB)
    return img


def scale_down(
        img: np.ndarray, *,
        width: int,
        height: int) -> np.ndarray:
    """Resize numpy array image to specified size [height, width]"""
    assert img.ndim == 3 and img.shape[-1] == 3, 'Expecting HWC image'

    # Resize using OpenCV
    return cv2.resize(
        img,
        (width, height),
        interpolation=cv2.INTER_AREA)


def load_tiles(
        paths: Iterable[Path], *,
        tile_width: int,
        tile_height: int):

    dbg = Path('dbg')
    dbg.mkdir(parents=True, exist_ok=True)

    for path in paths:
        tile = read_image_lab(path)
        yield scale_down(tile, width=tile_width, height=tile_height)


def write_jpeg(img: np.ndarray, path: str | Path, quality: int = 90):
    """Write numpy array image to JPEG file"""
    assert img.shape[-1] == 3, 'Expecting LAB image'

    img = cv2.cvtColor(img, cv2.COLOR_LAB2RGB)
    pil_img = Image.fromarray(img)
    pil_img.save(path, 'JPEG', quality=quality)



import json
from math import ceil, sqrt
from pathlib import Path
from typing import Iterable

import cv2
import numpy as np
from cattrs import structure
from tqdm import tqdm

from mozyq.io import read_image_lab, write_jpeg
from mozyq.mzq import Mozyq
from mozyq.util import center_crop


def crop_zoom(master: np.ndarray, zoom: float):
    if (zoom == 1):
        return master

    assert zoom > 1, 'Zoom must be greater than 1'

    h, w, _ = master.shape

    # Calculate crop dimensions (smaller than original)
    crop_h = int(h / zoom)
    crop_w = int(w / zoom)

    crop_h += crop_h % 2  # Make even
    crop_w += crop_w % 2  # Make even

    # Calculate crop offsets (center crop)
    i = (h - crop_h) // 2
    j = (w - crop_w) // 2

    # First: Crop to smaller size
    cropped = master[i:i + crop_h, j:j + crop_w]

    # Then: Zoom (resize) back to original size
    zoomed = cv2.resize(
        cropped,
        (w, h),
        interpolation=cv2.INTER_LANCZOS4)

    zh, zw, _ = zoomed.shape
    assert (zh == h) and (zw == w), 'Error in zoomed dimensions'
    return zoomed


def smart_scale_down_crop(
        grid: np.ndarray, *,
        grid_height: int,
        grid_width: int,
        crop_height: int,
        crop_width: int):

    out_height = crop_height
    out_width = crop_width
    grid_height += grid_height % 2  # Make even
    grid_width += grid_width % 2  # Make even

    h, *_ = grid.shape
    scale = grid_height / h

    if scale < .5:
        grid = cv2.resize(
            grid,
            (grid_width, grid_height),
            interpolation=cv2.INTER_AREA)

    else:
        crop_height = ceil(crop_height / scale)
        crop_width = ceil(crop_width / scale)

    crop = center_crop(
        img=grid,
        height=crop_height,
        width=crop_width)

    if scale >= .5:
        crop = cv2.resize(
            crop,
            (out_width, out_height),
            interpolation=cv2.INTER_AREA)

    return grid, crop


def ease(s: float, e: float, n: int = 90):
    """Generate n values from s to e with ease-out effect"""
    t = np.linspace(0, 1, n)
    t = np.where(
        t < 0.5,
        4 * t**3,
        1 - (-2 * t + 2)**3 / 2)

    return s + (e - s) * t


def step(
        *,
        master: np.ndarray,
        tiles: list[np.ndarray],
        zooms: Iterable[float]):

    d = int(sqrt(len(tiles)))
    assert d ** 2 == len(tiles), 'Tiles length must be a perfect square'

    grid = [[tiles[i * d + j]
             for j in range(d)]
            for i in range(d)]

    rows = [np.hstack(row) for row in grid]
    grid = np.vstack(rows)

    h, w, _ = grid.shape
    crop_height = h // d
    crop_width = w // d

    max_zoom = max(zooms)
    for zoom in zooms:
        alpha = (1 / zoom) * 0.3
        target = crop_zoom(master, zoom)
        scale = zoom / max_zoom
        grid, crop = smart_scale_down_crop(
            grid,
            grid_height=ceil(h * scale),
            grid_width=ceil(w * scale),
            crop_height=crop_height,
            crop_width=crop_width)

        yield (1 - alpha) * crop + alpha * target

    for beta in np.linspace(alpha, 1, 30):
        yield (1 - beta) * crop + beta * target


def frames(mzq: list[Mozyq], out_folder: Path):
    out_folder.mkdir(parents=True, exist_ok=True)

    i = 0
    for mozyq in mzq:
        max_zoom = int(sqrt(len(mozyq.tiles)))
        zooms = ease(max_zoom, 1)

        master = read_image_lab(mozyq.master)
        tiles = [
            read_image_lab(tile_path)
            for tile_path in mozyq.tiles]

        fs = step(
            master=master,
            tiles=tiles,
            zooms=zooms)

        for frame in tqdm(fs):
            write_jpeg(
                frame.astype(np.uint8),
                out_folder / f'{i:04d}.jpg')

            i += 1


def read_all(mzq: list[Mozyq]):
    for mozyq in mzq:
        yield read_image_lab(mozyq.master)
        for tile_path in mozyq.tiles:
            yield read_image_lab(tile_path)


if __name__ == '__main__':
    with open('output.json') as f:
        mzqs = [
            structure(mzq, Mozyq)
            for mzq in json.load(f)]

    frames(mzqs, Path('./frames'))



from contextlib import contextmanager
from time import perf_counter

import numpy as np


@contextmanager
def timer(label=""):
    start = perf_counter()
    yield
    end = perf_counter()
    print(f"{label} took {(end - start)*1000:.2f} ms")


def center_crop(
        *,
        img: np.ndarray,
        height: int,
        width: int):

    height -= height % 2  # Make even
    width -= width % 2  # Make even
    h, w, _ = img.shape

    assert (height <= h) and (width <= w), \
        f'Crop size must be <= image size: {height}x{width} <= {h}x{w}'

    i = (h - height) // 2
    j = (w - width) // 2
    return img[i:i + height, j:j + width]



import json
from pathlib import Path

import numpy as np
from cattr import unstructure
from scipy.optimize import linear_sum_assignment
from scipy.spatial.distance import cdist
from skimage.util import view_as_blocks
from tqdm import tqdm

from mozyq.io import load_tiles, read_image_lab, write_jpeg
from mozyq.mozyq_types import Mozyq


class MozyqGenerator:
    def __init__(
            self, *,
            paths: list[Path],
            vecs: np.ndarray,
            tile_width: int,
            tile_height: int,
    ):

        assert vecs.ndim == 2, f'vectors must be 2D {vecs.shape}'

        _, s = vecs.shape

        vec_size = tile_width * tile_height * 3
        assert s == vec_size, \
            f'vectors must be of size {vec_size}'

        self.paths = np.array(paths)
        self.vecs = vecs
        self.tile_width = tile_width
        self.tile_height = tile_height

    @classmethod
    def from_folder(
            cls, folder: Path, *,
            tile_width: int,
            tile_height: int,):

        ps = sorted(list(folder.glob('*.jpg')))

        tiles = load_tiles(
            tqdm(ps, desc='reading tiles'),
            tile_width=tile_width,
            tile_height=tile_height)

        vecs = [
            tile.ravel().astype(np.float32)
            for tile in tqdm(tiles, desc='vectorizing tiles')]

        vecs = np.stack(vecs)

        return cls(
            paths=ps,
            vecs=vecs,
            tile_width=tile_width,
            tile_height=tile_height)

    def generate(self, master: np.ndarray):
        h, w, c = master.shape

        assert c == 3, 'master image must be LAB'
        assert h % 2 == 0, 'master image height must be even'
        assert w % 2 == 0, 'master image width must be even'
        assert master.size <= self.vecs.size, 'master image too large'

        master = master.astype(np.float32)

        targets = view_as_blocks(
            master,
            block_shape=(self.tile_height, self.tile_width, 3)
        ).reshape(-1, self.tile_height * self.tile_width * 3)

        def dist(tile, patch):
            return np.linalg.norm(tile - patch)

        # Compute distance matrix using scipy
        d = cdist(self.vecs, targets, metric=dist)
        rid, cid = linear_sum_assignment(d)

        # Sort indices
        return self.paths[rid][np.argsort(cid)]


def gen_mzq_json(
        *,
        master: Path,
        tile_folder: Path,
        width: int,
        height: int,
        num_tiles: int,
        output_json: Path,
        max_transitions: int = 50,):

    assert num_tiles % 2 == 1, 'num_tiles must be odd'
    assert width % 2 == 0, 'width must be even'
    assert height % 2 == 0, 'height must be even'
    assert width % num_tiles == 0, 'width must be divisible by num_tiles'
    assert height % num_tiles == 0, 'height must be divisible by num_tiles'

    tile_width = width // num_tiles
    tile_height = height // num_tiles

    gen = MozyqGenerator.from_folder(
        tile_folder,
        tile_width=tile_width,
        tile_height=tile_height)

    masters = set()
    mzqs: list[Mozyq] = []
    for _ in tqdm(range(max_transitions)):
        if master in masters:
            print(f'Master {master} already used, stopping generation')
            break

        masters.add(master)

        # GENERATE
        paths = gen.generate(read_image_lab(master))
        mzqs.append(Mozyq(master=master, tiles=paths.tolist()))

        # Use center tile as next master
        master = paths[len(paths) // 2]

    # WRITE JSON
    with output_json.open('w') as f:
        json.dump(unstructure(mzqs[:-1][::-1]), f)

    print(f'Wrote Mozyq JSON to {output_json}')


def blocks(
        img_jpg: Path, *,
        num_blocks: int = 5,
        output_folder: Path = Path('./blocks')):

    output_folder.mkdir(parents=True, exist_ok=True)

    img = read_image_lab(img_jpg)
    h, w, c = img.shape
    assert c == 3, 'image must be LAB'

    blocks = view_as_blocks(
        img, block_shape=(h // num_blocks, w // num_blocks, 3))

    for i, block in enumerate(blocks.reshape(-1, h // num_blocks, w // num_blocks, 3)):
        block_path = output_folder / f'block_{i:03d}.jpg'
        write_jpeg(block.astype(np.uint8), str(block_path))


if __name__ == '__main__':
    master = Path('./normalized/0518.jpg')
    tile_folder = Path('./blocks')

    gen_mzq_json(
        master=master,
        tile_folder=master.parent,
        width=600,
        height=750,
        num_tiles=15,
        output_json=Path('./output.json')
    )



from pathlib import Path

from PIL import Image
from tqdm import tqdm


def normalize(
        *,
        in_folder: Path,
        out_folder: Path,
        min_width: int,
        min_height: int,
        target_width: int,
        target_height: int):

    out_folder.mkdir(parents=True, exist_ok=True)

    image_extensions = {'.jpg', '.jpeg', '.png',
                        '.bmp', '.tiff', '.tif', '.webp'}
    image_files = [
        f for f in in_folder.iterdir()
        if f.is_file() and f.suffix.lower() in image_extensions
    ]

    processed_count = 0
    skipped_count = 0

    for i, image_path in enumerate(tqdm(image_files, desc="Processing images")):
        try:
            with Image.open(image_path) as img:
                width, height = img.size

                if width < min_width or height < min_height:
                    skipped_count += 1
                    continue

                if img.mode != 'RGB':
                    img = img.convert('RGB')

                aspect_ratio = height / width
                new_height = int(target_width * aspect_ratio)
                img_resized = img.resize(
                    (target_width, new_height),
                    Image.Resampling.LANCZOS)

                if new_height >= target_height:
                    top = (new_height - target_height) // 2
                    bottom = top + target_height
                    img_cropped = img_resized.crop(
                        (0, top, target_width, bottom))
                else:
                    skipped_count += 1
                    continue

                output_path = out_folder / f'{i:04d}.jpg'
                img_cropped.save(output_path, 'JPEG', quality=95)
                processed_count += 1

        except Exception as e:
            print(f"Error processing {image_path}: {e}")
            skipped_count += 1
            continue

    print(
        f"Processing complete: {processed_count} images processed, {skipped_count} images skipped")
